#include "cTree.h"
#include "macros.h"

#include <cmath>
using namespace std;

#define ran(seed) (seed = ((seed * 1234567) % 65535))


int drawTreePart(int depth, int maxdepth, float length, int seed, GLuint trunk_displaylist, GLuint leaf_displaylist)
{
  // Recursion end?
  if (depth >= maxdepth) return seed;
  //
  const float height = length*(maxdepth - depth);
  const float width = 0.035*length+(maxdepth - depth)*(maxdepth - depth)*length*0.0125f;
  bool leaves = true;
  // Draw Trunk and leaf if at max depth.
  glPushMatrix();
      glScalef(width*0.5f, height*0.5f, width*0.5f);
      glTranslatef(0, 1, 0);
      glCallList(trunk_displaylist);
      if ((depth == maxdepth-1) && leaves) {
        const float s = 3.6;
        //glTranslatef(0, 0.5f, 0);
        glScalef(s*height/width, s*0.5, s*height/width);
        glTranslatef(0, 0.5, 0);
        glCallList(leaf_displaylist);
      }
  glPopMatrix();
  // Draw Offsprings
  
  const int n = 3;
  int rnums[2*n];
  int r = seed;
  int j = 0;
  for (int i = 0; i < n; i++)
  {
      rnums[j++] = ran(r);
      rnums[j++] = ran(r);
  }
  
  int anglex, angley;
  j = 0;
  for (int i = 0; i < n; i++)
  {
    const int yvar = 15;
    angley = int(i*360.0f/float(n)) - yvar + rnums[j++]%(2*yvar);
    const float xvar = 27.0f;
    anglex = int(60.0f - xvar + (rnums[j++] % 100)*0.01f * (2.0f*xvar)*(1.0f-float(maxdepth-depth)/float(maxdepth)));
    glPushMatrix();
        glTranslatef(0, (i + 2)*height/float(n + 1)-width*0.5f, 0);
        glRotatef(angley, 0, 1, 0);
        glRotatef(anglex, 1, 0, 0);
        r = drawTreePart(depth+1, maxdepth, length, r, trunk_displaylist, leaf_displaylist);
    glPopMatrix();
  };
  return r;
}


void drawRubberTreeLeaf()
{
    glPushAttrib(GL_ENABLE_BIT);
    glDisable(GL_CULL_FACE);
    glBegin(GL_TRIANGLE_FAN);
        // 0 back
        glNormal3f(0,+1,0);
        glColor3f(0.2f, 0.1f, 0.0f);
        glVertex3f( 0.0,   0.0,   0.0);
        // 1 right
        glNormal3f(-1,+1,0);
        glColor3f(0.0f, 0.4f, 0.0f);
        glVertex3f(+0.3,  +0.1,  +0.1);
        // 2 right
        glColor3f(0.0f, 0.3f, 0.0f);
        glVertex3f(+0.4,  +0.2,  +0.2);
        // 3 right
        glColor3f(0.0f, 0.4f, 0.0f);
        glVertex3f(+0.35,  +0.2,  +0.5);
        // 4 right
        glColor3f(0.0f, 0.3f, 0.0f);
        glVertex3f(+0.15,  +0.1,  +0.8);
        // 5 tip
        glNormal3f(+1,+1,0);
        glColor3f(0.0f, 0.2f, 0.0f);
        glVertex3f( 0.0,   0.0,  +1.1);
        // 4 left
        glColor3f(0.0f, 0.3f, 0.0f);
        glVertex3f(-0.15,  +0.1,  +0.8);
        // 3 left
        glColor3f(0.0f, 0.4f, 0.0f);
        glVertex3f(-0.35,  +0.2,  +0.5);
        // 2 left
        glColor3f(0.0f, 0.3f, 0.0f);
        glVertex3f(-0.4,  +0.2,  +0.2);
        // 1 left
        glColor3f(0.0f, 0.4f, 0.0f);
        glVertex3f(-0.3,  +0.1,  +0.0);
        // 0 back
        glNormal3f(0,+1,0);
        glColor3f(0.2f, 0.1f, 0.0f);
        glVertex3f( 0.0,   0.0,   0.0);
    glEnd();
    glPopAttrib();
}


void drawCaribeanTreeLeaf()
{
    glPushAttrib(GL_ENABLE_BIT);
    glDisable(GL_CULL_FACE);
    const float step = 6 * M_PI / 180.0f;
    float alpha = step*0.5;
    glBegin(GL_TRIANGLES);
    for (int i = 0; i < 5; i++) {
        float l[] = { 1.0, 0.7, 0.5, 0.3, 0.2 };
        //float length = 1-sin(10*i*0.6*i * 0.017454); //3.0 / (1.0f + i);
        //length *= 3;
        float length = 2.5*l[i];
        float ulength = 1.7 * length;
        //printf("len %i = %f\n", i, length);
        float tips = 0.7;
        // 0 back
        glNormal3f(0,+1,0);
        glColor3f(0.0f, 0.3f, 0.0f);
        glVertex3f( 0.0,   0.0,   0.0);
        glColor3f(0.5f, 0.4f, 0.2f);
        // 1
        glNormal3f(0,+1,0);
        glVertex3f( length*sin(alpha),   ulength*cos(alpha),   0.0);
        // 2
        glNormal3f(0,+1,0);
        glVertex3f( tips*length*sin(alpha+step),   tips*ulength*cos(alpha+step),   0.0);
        // 0 back
        glNormal3f(0,+1,0);
        glColor3f(0.0f, 0.3f, 0.0f);
        glVertex3f( 0.0,   0.0,   0.0);
        glColor3f(0.5f, 0.4f, 0.2f);
        // 1
        glNormal3f(0,+1,0);
        glVertex3f( length*sin(-alpha),   ulength*cos(-alpha),   0.0);
        // 2
        glNormal3f(0,+1,0);
        glVertex3f( tips*length*sin(-alpha-step),   tips*ulength*cos(-alpha-step),   0.0);
        //
        alpha += 2*step;
    }
    glEnd();
    glPopAttrib();
}


void drawHalmTreeLeaf()
{
    glPushMatrix();
    glScalef(0.2, 0.7, 2.5);
        drawRubberTreeLeaf();
    glPopMatrix();
}


void drawButterflyTreeLeaf()
{
    glPushMatrix();
        glRotatef(-30, 0,1,0);
        drawHalmTreeLeaf();
        glRotatef(30, 0,1,0);
        drawHalmTreeLeaf();
        glRotatef(30, 0,1,0);
        drawHalmTreeLeaf();
    glPopMatrix();
}


#include <list>
using namespace std;

struct tree {
    int seed;
    int type;
    int age;
    GLuint list;
};
static list<tree*> cache;

static GLuint compileTreeDisplaylist(int seed, int type, int age)
{
    for (list<tree*>::iterator i = cache.begin(); i != cache.end(); i++) {
        tree* t = *i;
        if (t->seed == seed && t->type == type && t->age == age) {
            //printf("found cached tree seed: %i type: %i\n", seed, type);
            return t->list;
        }
    }
    
    printf("tree chache size: %i\n", cache.size());
    printf("tree not cached - generating\n");

    static GLint leaf_displaylists[] = { -1, -1, -1, -1, -1};
    if (leaf_displaylists[type] == -1) {
        leaf_displaylists[type] = glGenLists(1);
        glNewList(leaf_displaylists[type], GL_COMPILE);
            switch (type) {
                case 0: drawRubberTreeLeaf(); break;
                case 1: drawCaribeanTreeLeaf(); break;
                case 2: drawHalmTreeLeaf(); break;
                case 3: drawButterflyTreeLeaf(); break;
            };
        glEndList();
    }
    
    static GLuint trunk_displaylist = -1;
    if (trunk_displaylist == -1) {
        trunk_displaylist = glGenLists(1);
        glNewList(trunk_displaylist, GL_COMPILE);
            glColor3f(0.1f, 0.1f, 0.0f);
            glCenterUnitBlock();
        glEndList();
    }

    tree* t = new tree();
    t->seed = seed;
    t->type = type;
    t->age = age;
    t->list = glGenLists(1);
    glNewList(t->list, GL_COMPILE);
        float s = 2.5f + age*0.3;
        glScalef(s,s,s);
        drawTreePart(0, age, 0.3f+age*0.1, t->seed, trunk_displaylist, leaf_displaylists[type]);
    glEndList();
    
    cache.push_front(t);
    printf("generated\n");
    return t->list;
}


cTree::cTree(float* pos, float* rot, int seed, int type, int age)
{
    this->mRoles |= (1UL << cObject::TREE);
    memcpy(mPos, pos, sizeof(float)*3);
    memcpy(mRot, rot, sizeof(float)*3);
    this->seed = seed;
    this->type = type;
    this->age = age;
    this->list = -1;
}


void cTree::drawSolid()
{
    if (list == -1) list = compileTreeDisplaylist(this->seed, this->type, this->age);
    glPushAttrib(GL_ENABLE_BIT);
    glEnable(GL_NORMALIZE);
    glDisable(GL_TEXTURE_2D);
    glPushMatrix();
        glTranslatef(mPos[0], mPos[1], mPos[2]);
        glCallList(list);
    glPopMatrix();
    glPopAttrib();
}


