// cWorld.cpp

#include "cWorld.h"

#include "cBackground.h"
#include "cObject.h"
#include "cMission.h"

#include <GL/gl.h>

#define loopObjects(objset)   for (std::list<cObject *>::iterator it = (objset).begin(); it != (objset).end(); it++)
#define loopObjectsNoInc(objset)   for (std::list<cObject *>::iterator it = (objset).begin(); it != (objset).end();)


cWorld* cWorld::instance = NULL;


cWorld::cWorld() {
    cWorld::instance = this;
    
    mBackground = NULL;
    mMission = NULL;
    
    mGravity[0] = 0;
    mGravity[1] = 9.8f;
    mGravity[2] = 0;
    
    mFPS = 30.0f;
    mSPF = 1.0f / mFPS;
    mTime = 12.00f;
    mSerial = 1;
    //getSystemHour(mTime);
}


// Spawning And Fragging


void cWorld::spawnObject(cObject *object) {
    tSerial hh = (int) this->mTime;
    tSerial fff = (int) ((this->mTime - hh) * 1000);
    tSerial yyyy = 2028;
    tSerial mm = 8;
    tSerial dd = 9;
    tSerial serid = ((((yyyy * 100 + mm) * 100  + dd) * 100 + hh) * 100 + fff) * 100000 + mSerial;
        
    if (mIndex[serid] != NULL) {
        mSerial++;
        serid++;
    } else {
        mSerial = 1;
    }
    
    printf("%lli  %li\n", serid, mSerial);
    mIndex[serid] = object;
    object->mSerial = serid;
            
    mObjects.push_back(object);
    object->onSpawn();
}


void cWorld::fragObject(cObject *object) {
    if (object->hasRole(cObject::FRAGGED)) return;
    object->addRole(cObject::FRAGGED);
    mObjects.remove(object);
    mCorpses.push_back(object);
    mIndex.erase(object->mSerial);
    object->onFrag();
}


void cWorld::bagFragged() {
    // Finally delete objects which where fragged before.
    while (!mCorpses.empty()) {
        cObject* object = mCorpses.front();
        mCorpses.pop_front();
        delete object;
    };
}


// Simulation Step


void cWorld::advanceTime(int deltamsec) {
    if (mMission) mMission->checkConditions();
    mSPF = float(deltamsec) / 1000.0f;
    mFPS = 1.0 / (mSPF + 0.000001f);
    mTime += mSPF / 3600.0f;
    //mTime += mSPF; // speed up daycycle
    //printf("time: %.6f\n", mTime);
}


void cWorld::animateObjects() {
    float spf = mSPF;
    loopObjectsNoInc(mObjects) {
        cObject* object = *it++;
        object->animate(spf);
    };
}


void cWorld::transformObjects() {
    loopObjectsNoInc(mObjects) {
        cObject* object = *it++;
        glPushMatrix();
        object->transform();
        glPopMatrix();
    };
}


void cWorld::drawBack() {
    if (mBackground) mBackground->drawBackground(mTime);
}

void cWorld::drawSolid() {
    loopObjectsNoInc(mObjects) {
        cObject* object = *it++;
        glPushMatrix();
        object->drawSolid();
        glPopMatrix();
    };
}


void cWorld::drawEffect() {
    loopObjectsNoInc(mObjects) {
        cObject* object = *it++;
        glPushMatrix();
        object->drawEffect();
        glPopMatrix();
    };
}


// World Scanning And Filtering


void cWorld::printNames(std::list<cObject*>* objects) {
    char buffer[1024];
    std::string s = "[";
    if (objects == NULL) objects = &mObjects;
    loopObjects(*objects) {
        sprintf(buffer, "%llu", (*it)->mSerial);
        s += " ";
        s += ( (*it)->mName );
        s += "#";
        s += buffer;
        s += "(";
        sprintf(buffer, "%lu", (*it)->mRoles);
        s += buffer;
        s += ")";
    }
    s += " ]";
    printf("%s\n", s.c_str());
}


std::list<cObject*>* cWorld::filterByRole(cObject* ex, unsigned long rolemask, bool all, int maxamount, std::list<cObject*>* objects) {
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    if (objects == NULL) objects = &mObjects;
    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object->mRoles & (1UL << cObject::FRAGGED)) continue;
        // Filter Condition
        unsigned long intersection = object->mRoles & rolemask;
        if ((intersection != rolemask) && !((intersection != 0) && !all)) continue;
        if (object == ex) continue;
        amount--;
        result->push_back(object);
    }
    return result;
}


std::list<cObject*>* cWorld::filterByRange(cObject* ex, float* origin, float minrange, float maxrange, int maxamount, std::list<cObject*>* objects) {
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    if (objects == NULL) objects = &mObjects;
    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object == ex) continue;
        if (object->mRoles & (1UL << cObject::FRAGGED)) continue;
        // Filter Condition
        float d = lineDistance(origin, object->mPos);
        if ((d > maxrange+object->mRadius) || (d < minrange-object->mRadius)) continue;
        amount--;
        result->push_back(object);
    }
    return result;
}


std::list<cObject*>* cWorld::filterByName(cObject* ex, char* name, int maxamount, std::list<cObject*>* objects) {
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    if (objects == NULL) objects = &mObjects;
    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object == ex) continue;
        if (object->mRoles & (1UL << cObject::FRAGGED)) continue;
        // Filter Condition
        if (strcmp(object->mName.c_str(), name) != 0) continue;
        amount--;
        result->push_back(object);
    }
    return result;
}


std::list<cObject*>* cWorld::filterByBeam(cObject* ex, float* pointa, float* pointb, float radius, int maxamount, std::list<cObject*>* objects) {
    float* x1 = pointa;
    float* x2 = pointb;
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    if (objects == NULL) objects = &mObjects;
    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object == ex) continue;
        if (object->mRoles & (1UL << cObject::FRAGGED)) continue;
        // Filter Condition
        
        float* x0 = object->mPos;
        float u[] = { x0[0] - x1[0], x0[1] - x1[1], x0[2] - x1[2] };
        float v[] = { x2[0] - x1[0], x2[1] - x1[1], x2[2] - x1[2] };
        float a = u[0] * v[0] + u[1] * v[1] + u[2] * v[2];
        float b = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
        float delta = a/b;
        
        //printf("delta: %.2f\n", delta);
        
        if (delta < -0.01) continue;
        if (delta > +1.01) continue;
        
        float p[] = {
            x1[0] + delta * (x2[0] - x1[0]),
            x1[1] + delta * (x2[1] - x1[1]),
            x1[2] + delta * (x2[2] - x1[2])
        };
        
        float dv[] = {
            p[0] - x0[0],
            p[1] - x0[1],
            p[2] - x0[2]
        };
        
        float d2 = dv[0]*dv[0] + dv[1]*dv[1] + dv[2]*dv[2];
        float is = radius + object->mRadius;
        
        if (d2 > is*is) continue;
        
        amount--;
        result->push_back(object);
    }
    return result;
}


