/* 
 * File:   main.cpp
 * Author: Benjamin Pickhardt (hackCraft)
 *
 * Created on March 28, 2008, 21:25 PM (1999)
 */

#include "main.h"
#include <GL/glut.h>


cPad* gKeyPad = NULL; // Keyboard inputs send to which Gamepad?
cObject* gCameraObject = NULL;
cWorld *planet = NULL;

bool external_camera = false;
bool satelite_camera = false;
bool paused          = false;
bool fullscreen      = false;
bool wireframe       = false;
bool nightvision     = false;


struct Mouse {
    Mouse() { memset(this, 0, sizeof(Mouse)); }
    unsigned short buttons;
    float speed[2];
    float pos[2];
} gMouse;


void initMission() {
  planet = new cWorld();  
  planet->mMission = new cTestMission();
  gCameraObject = planet->mMission->init(planet);
  gKeyPad = gCameraObject->mPad;
}


void initGL()
{
  float fogColor[4] = { 1.0, 1.0, 1.0, 1.0 };
  float fogDensity = 0.001;

  glEnable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);

  glClearColor(0.0, 0.125, 0.0, 0.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glShadeModel(GL_SMOOTH);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDepthMask(GL_TRUE);
  glDepthFunc(GL_LEQUAL);

  glFogi(GL_FOG_MODE, GL_EXP);
  glFogfv(GL_FOG_COLOR, fogColor);
  glFogf(GL_FOG_DENSITY, fogDensity);

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  // Lighting
  glEnable(GL_LIGHTING);
  float p[] = { 0, 1000, 0, 1 };
  glLightfv(GL_LIGHT0, GL_POSITION, p);
  glEnable(GL_COLOR_MATERIAL);
}


void Reshape(int w, int h)
{
  //glClearAccum(0.0, 0.0, 0.0, 0.0);
  glClear(GL_ACCUM_BUFFER_BIT);
  glViewport(0,0,w,h);
  glutPostRedisplay();
}


void Draw()
{
  // Calculate Timediffence between frames.
  static int lasttime = 0;
  int currenttime = glutGet(GLUT_ELAPSED_TIME);
  int elapsed = currenttime - lasttime;
  lasttime = currenttime;
  
  //if (hurlmotion) elapsed = int(elapsed * 1.0f);

  // calculate the time that the last frame took.
  planet->advanceTime(elapsed);

  // Delete Fragged Objects of previous frame.
  planet->bagFragged();

  // Calculate new animation state and the like (if not paused).
  if (!paused)
      planet->animateObjects();

  // Calculate positions through transformations.
  glPushMatrix();
    glLoadIdentity();
      planet->transformObjects();
  glPopMatrix();

  // SOLID / WIREFRAME drawing scheme.
  if (!wireframe) {
    float dark[] = { 0.0, 0.0, 0.0, 1.0 };
    glClearColor(0.2, 0.2, 0.2, 1.0);
    glEnable(GL_FOG);
    glFogfv(GL_FOG_COLOR, dark);
    //glClearColor(dome->mLoColor[X], dome->mLoColor[Y], dome->mLoColor[Z], 1.0);
    //glEnable(GL_FOG);
    //glFogfv(GL_FOG_COLOR, dome->mLoColor);
    glEnable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glEnable(GL_CULL_FACE);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
  } else {
    float dark[] = { 0.0, 0.0, 0.0, 1.0 };
    glClearColor(0.2, 0.2, 0.2, 1.0);
    glEnable(GL_FOG);
    glFogfv(GL_FOG_COLOR, dark);
    glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glDisable(GL_CULL_FACE);
  };

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glPushPerspectiveProjection(60);

  // Setup camera.
  cMech* mech = (cMech*) gCameraObject;
  float *matrixfv = mech->getPilotMatrix();
  assert(matrixfv != NULL);
  glLoadIdentity();
  if (external_camera) {
    glTranslatef(0, 0, -5);
    glRotatef(15, 1, 0, 0);
  } else if (satelite_camera) {
    glTranslatef(0, 0, -100);
    glRotatef(90, 1, 0, 0);
  };
  glRotatef(180, 0, 1, 0);
  glMultMatrixf(matrixfv);

  // Draw the surounding Sky- and Ground-Dome/Box
  if (planet->mBackground) planet->mBackground->drawBackground(planet->mTime);

  //glBindTexture(GL_TEXTURE_2D, gAttentionTextureBind);

  // Draw the Objects themselves.
  planet->drawSolid();

  // Draw the Object's translucent effects.
  planet->drawEffect();

  // Draw the Head-Up-Display of the currently spectating Object.
  gCameraObject->drawHUD();

  glPopProjection();

  glFlush();

  //const float motionblur = 0.73f;
  //glAccum(GL_MULT, motionblur); // accum = accum * factor
  //glAccum(GL_ACCUM, 1.0f-motionblur); // accum = accum + color * factor
  //glAccum(GL_RETURN, 1.0f); // color = accum * factor
  
  if (nightvision) {
    const float motionblur = 0.83f;
    glAccum(GL_MULT, motionblur); // accum = accum * factor
    glAccum(GL_ACCUM, 1.0f-motionblur); // accum = accum + color * factor
    glAccum(GL_MULT, -1.0f); glAccum(GL_ADD, 1.0f); // inversion
    glAccum(GL_RETURN, 1.0f); // color = accum * factor
    glAccum(GL_ADD, -1.0f); glAccum(GL_MULT, -1.0f); // inversion
  }
  
  glutSwapBuffers();
  
  //gPad1.print();
}


void Idle()
{
  glutPostRedisplay();
}


void MouseFunc(int button, int state, int x, int y)
{
  if (button == GLUT_RIGHT_BUTTON)   if (state == GLUT_DOWN) gMouse.buttons |= 2; else gMouse.buttons &= !2;
  if (button == GLUT_MIDDLE_BUTTON)  if (state == GLUT_DOWN) gMouse.buttons |= 4; else gMouse.buttons &= !4;
  if (button == GLUT_LEFT_BUTTON)    if (state == GLUT_DOWN) gMouse.buttons |= 1; else gMouse.buttons &= !1;
  GLint view[4];
  glGetIntegerv(GL_VIEWPORT, view);
  if (x > view[2]) x = view[2];
  if (y > view[3]) y = view[3];
  gMouse.pos[0] = x;
  gMouse.pos[1] = y;
  gMouse.speed[0] = (y-view[2]/2.)/(view[2]/4.);
  gMouse.speed[1] = (x-view[3]/2.)/(view[3]/4.);
}


void Mouse(int x, int y)
{
  GLint view[4];
  glGetIntegerv(GL_VIEWPORT, view);
  if (x > view[2]) x = view[2];
  if (y > view[3]) y = view[3];
  gMouse.pos[0] = x;
  gMouse.pos[1] = y;
  gMouse.speed[0] = (y-view[2]/2.)/(view[2]/4.);
  gMouse.speed[1] = (x-view[3]/2.)/(view[3]/4.);
}


void KeyDn(unsigned char k, int x, int y)
{
  if (k == 27) {
    exit(0);
  };

  if (k == CAMERA_KEY) {
    external_camera = !external_camera;
    satelite_camera = 0;
  } else if (k == SAT_CAMERA_KEY) {
    satelite_camera = !satelite_camera;
    external_camera = 0;
  } else if (k == WIREFRAME_KEY) {
    wireframe = !wireframe;
  } else if (k == NIGHTVISION_KEY) {
    nightvision = !nightvision;
  } else if (k == PAUSE_KEY) {
    if (!paused) {
      Draw();
    } else {
    };
    paused = !paused;
  };

  if (gKeyPad == NULL) return;
  
  int mod = glutGetModifiers();
  
  if (mod != GLUT_ACTIVE_ALT) {
    if (k == 'j') gKeyPad->setAxis(cPad::AX_LR2, -1.0f);
    if (k == 'l') gKeyPad->setAxis(cPad::AX_LR2, +1.0f);
    if (k == 'i') gKeyPad->setAxis(cPad::AX_UD2, -1.0f);
    if (k == 'k') gKeyPad->setAxis(cPad::AX_UD2, +1.0f);
  } else if (mod == GLUT_ACTIVE_ALT) {
    if (k == 'j') gKeyPad->setButton(cPad::BT_HL, true);
    if (k == 'l') gKeyPad->setButton(cPad::BT_HR, true);
    if (k == 'i') gKeyPad->setButton(cPad::BT_HU, true);
    if (k == 'k') gKeyPad->setButton(cPad::BT_HD, true);
  };
  
  if (k == 'a') gKeyPad->setButton(cPad::BT_L1, true);
  if (k == 'g') gKeyPad->setButton(cPad::BT_R1, true);
  
  if (k == 'w') gKeyPad->setButton(cPad::BT_L2, true);
  if (k == 'r') gKeyPad->setButton(cPad::BT_R2, true);
  
  if (k == 's') gKeyPad->setButton(cPad::BT_SQUARE, true);
  if (k == 'f') gKeyPad->setButton(cPad::BT_CIRCLE, true);
  
  if (k == 'e') gKeyPad->setButton(cPad::BT_DELTA, true);
  if (k == 'd') gKeyPad->setButton(cPad::BT_CROSS, true);
  
  if (k == '.') gKeyPad->setButton(cPad::BT_J1B, true);
  if (k == ',') gKeyPad->setButton(cPad::BT_J2B, true);
}

void KeyUp(unsigned char k, int x, int y)
{
  if (gKeyPad == NULL) return;
  if (k == 'j') gKeyPad->setAxis(cPad::AX_LR2, 0.0f);
  if (k == 'l') gKeyPad->setAxis(cPad::AX_LR2, 0.0f);
  if (k == 'i') gKeyPad->setAxis(cPad::AX_UD2, 0.0f);
  if (k == 'k') gKeyPad->setAxis(cPad::AX_UD2, 0.0f);
  
  if (k == 'j') gKeyPad->setButton(cPad::BT_HL, false);
  if (k == 'l') gKeyPad->setButton(cPad::BT_HR, false);
  if (k == 'i') gKeyPad->setButton(cPad::BT_HU, false);
  if (k == 'k') gKeyPad->setButton(cPad::BT_HD, false);
  
  if (k == 'a') gKeyPad->setButton(cPad::BT_L1, false);
  if (k == 'g') gKeyPad->setButton(cPad::BT_R1, false);
  
  if (k == 'w') gKeyPad->setButton(cPad::BT_L2, false);
  if (k == 'r') gKeyPad->setButton(cPad::BT_R2, false);
  
  if (k == 's') gKeyPad->setButton(cPad::BT_SQUARE, false);
  if (k == 'f') gKeyPad->setButton(cPad::BT_CIRCLE, false);
  
  if (k == 'e') gKeyPad->setButton(cPad::BT_DELTA, false);
  if (k == 'd') gKeyPad->setButton(cPad::BT_CROSS, false);
  
  if (k == '.') gKeyPad->setButton(cPad::BT_J1B, false);
  if (k == ',') gKeyPad->setButton(cPad::BT_J2B, false);
}


void SpecialDn(int k, int x, int y)
{
  if (gKeyPad == NULL) return;
  if (k == GLUT_KEY_LEFT) gKeyPad->setAxis(cPad::AX_LR1, -1.0f);
  if (k == GLUT_KEY_RIGHT) gKeyPad->setAxis(cPad::AX_LR1, +1.0f);
  if (k == GLUT_KEY_UP) gKeyPad->setAxis(cPad::AX_UD1, -1.0f);
  if (k == GLUT_KEY_DOWN) gKeyPad->setAxis(cPad::AX_UD1, +1.0f);
}

void SpecialUp(int k, int x, int y)
{
  if (gKeyPad == NULL) return;
  if (k == GLUT_KEY_LEFT) gKeyPad->setAxis(cPad::AX_LR1, 0.0f);
  if (k == GLUT_KEY_RIGHT) gKeyPad->setAxis(cPad::AX_LR1, 0.0f);
  if (k == GLUT_KEY_UP) gKeyPad->setAxis(cPad::AX_UD1, 0.0f);
  if (k == GLUT_KEY_DOWN) gKeyPad->setAxis(cPad::AX_UD1, 0.0f);
}

//
// GOD begins below!
//

#ifdef _WIN32

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  int    argc = 2;
  char  *argv[] = { "Nothing", "Nothing", "Nothing" };

#else

int main(int argc, char **argv) {

#endif

  int width = 640;
  int height = 480;

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB | GLUT_ACCUM);  
  glutInitWindowSize(width, height);
  glutInitWindowPosition(0, 0);
  glutCreateWindow("LinWarrior 3D - v. 0.61 public beta");

  glutDisplayFunc(*Draw);
  glutReshapeFunc(*Reshape);
  glutIdleFunc(*Idle);
  
  glutMouseFunc(*MouseFunc);
  glutPassiveMotionFunc(*Mouse);
  glutMotionFunc(*Mouse);
  
  glutKeyboardFunc(*KeyDn);
  glutKeyboardUpFunc(*KeyUp);
  glutSpecialFunc(*SpecialDn);
  glutSpecialUpFunc(*SpecialUp);

  glutSetCursor(GLUT_CURSOR_NONE);
  glutSetKeyRepeat(GLUT_KEY_REPEAT_OFF);

  initGL();
  initMission();
  glutMainLoop();

  return 0;
}

